from logging import Logger

from rich.spinner import Spinner

from prfiesta import SPINNER_STYLE


def update_spinner(message: str, spinner: Spinner, logger: Logger) -> None:
    logger.debug(message)
    if spinner:
        spinner.update(text=message, style=SPINNER_STYLE)



import logging
from datetime import datetime

import click
import cloup
from rich.live import Live
from rich.spinner import Spinner
from rich.text import Text

from prfiesta import SPINNER_STYLE, __version__
from prfiesta.collectors.github import GitHubCollector
from prfiesta.environment import GitHubEnvironment
from prfiesta.output import output_frame

logger = logging.getLogger(__name__)

github_environment = GitHubEnvironment()


@cloup.command()
@cloup.option_group(
    "general options",
    cloup.option("-u", "--users", multiple=True, help="The GitHub Users to search for. Can be multiple"),
    cloup.option("--reference", help="A external ticket reference to search for e.g. JIRA-1234"),
)
@cloup.option_group(
    "date filter options",
    cloup.option("-a", "--after", type=click.DateTime(formats=["%Y-%m-%d"]), help="Only search for pull requests after this date e.g 2023-01-01"),
    cloup.option("-b", "--before", type=click.DateTime(formats=["%Y-%m-%d"]), help="Only search for pull requests before this date e.g 2023-04-30"),
    cloup.option("-d", "--use-updated", is_flag=True, default=False, help="filter on when the pr was last updated rather then created"),
)
@cloup.option_group(
    "user filter options",
    cloup.option("-i", "--use-involves", is_flag=True, default=False, help="collect prs where the users are the author or assignee or mentioned or commented"),
    cloup.option("-r", "--use-reviewed-by", is_flag=True, default=False, help="collect prs where the users reviewed them"),
    cloup.option("-rr", "--use-review-requested", is_flag=True, default=False, help="collect prs where the users were requested a review"),
    constraint=cloup.constraints.mutually_exclusive,
    help="Collect alternative details for the users. If omitted then just collect the prs that a user has authored.",
)
@cloup.option_group(
    "output options",
    cloup.option("-o", "--output", default=None, help="The output location"),
    cloup.option(
        "-ot",
        "--output-type",
        type=click.Choice(["csv", "parquet", "duckdb"]),
        default="csv",
        show_default=True,
        show_choices=True,
        help="The output format",
    ),
    cloup.option("-dc", "--drop-columns", multiple=True, help="Drop columns from the output dataframe"),
)
@cloup.option_group(
    "authentication options",
    cloup.option("-x", "--url", help="The URL of the Git provider to use"),
    cloup.option("-t", "--token", help="The Authentication token to use"),
)
@cloup.version_option(__version__)
def main(**kwargs) -> None:
    users: tuple[str] = kwargs.get("users")
    token: str = kwargs.get("token") or github_environment.get_token()
    url: str = kwargs.get("url") or github_environment.get_url()
    output: str = kwargs.get("output")
    output_type: str = kwargs.get("output_type")
    before: datetime = kwargs.get("before")
    after: datetime = kwargs.get("after")
    drop_columns: list[str] = list(kwargs.get("drop_columns"))
    use_updated: bool = kwargs.get("use_updated")
    use_involves: bool = kwargs.get("use_involves")
    use_reviewed_by: bool = kwargs.get("use_reviewed_by")
    use_review_requested: bool = kwargs.get("use_review_requested")
    reference: str = kwargs.get("reference")

    logger.info("[bold green]PR Fiesta 🦜🥳")

    spinner = Spinner("dots", text=Text("Loading", style=SPINNER_STYLE))

    with Live(spinner, refresh_per_second=20, transient=True):
        collector = GitHubCollector(token=token, url=url, spinner=spinner, drop_columns=drop_columns)
        pr_frame = collector.collect(
            *users,
            after=after,
            before=before,
            use_updated=use_updated,
            use_involves=use_involves,
            use_reviewed_by=use_reviewed_by,
            use_review_requested=use_review_requested,
            reference=reference,
        )

        if not pr_frame.empty:
            logger.info("Found [bold green]%s[/bold green] pull requests!", pr_frame.shape[0])

            output_frame(pr_frame, output_type, spinner=spinner, output_name=output)
            logger.info("Time to analyze 🔎 See https://github.com/kiran94/prfiesta/blob/main/docs/analysis.md for some inspiration!")


if __name__ == "__main__":  # pragma: nocover
    main()



import importlib.metadata
import logging
import os

from rich.logging import RichHandler

__version__ = importlib.metadata.version(__name__)

LOGGING_LEVEL = os.environ.get("LOGGING_LEVEL", logging.INFO)
LOGGING_FORMAT = os.environ.get("LOGGING_FORMAT", "%(message)s")
SPINNER_STYLE = os.environ.get("SPINNER_STYLE", "blue")

logging.basicConfig(
    level=LOGGING_LEVEL,
    format=LOGGING_FORMAT,
    handlers=[RichHandler(markup=True, show_path=False, show_time=False, show_level=True)],
)



import calendar
import logging
from typing import Optional, Union

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes

logger = logging.getLogger(__name__)

_months = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]


def plot_state_distribution(data: pd.DataFrame, **kwargs) -> Union[plt.Figure, plt.Axes, pd.DataFrame]:
    ax: Optional[Axes] = kwargs.get("ax")
    palette: Optional[str] = kwargs.get("palette")
    title: Optional[str] = kwargs.get("title", "State Distribution")
    hue: Optional[str] = kwargs.get("hue", "repository_name")

    if ax:
        ax.set_title(title)

    return sns.histplot(data, x="state", hue=hue, ax=ax, palette=palette)


def plot_overall_timeline(data: pd.DataFrame, **kwargs) -> Union[plt.Figure, plt.Axes, pd.DataFrame]:
    ax: Optional[Axes] = kwargs.get("ax")
    palette: Optional[str] = kwargs.get("palette")
    title: Optional[str] = kwargs.get("title", "Overall Contributions")
    hue: Optional[str] = kwargs.get("hue", "month")

    temp = data.copy()

    temp["month"] = temp["created_at"].dt.month_name()
    temp["year"] = temp["created_at"].dt.year.astype(str)

    temp = temp.groupby(["month", "year"])["id"].count()
    temp = temp.reset_index()

    # X Axis Ordering (Year)
    x_order = temp["year"].unique().tolist()
    x_order = [int(x) for x in x_order]
    x_order.sort()
    x_order = [str(x) for x in x_order]

    # Hue Ordering (Months)
    sorted_months = sorted(_months, key=lambda x: list(calendar.month_name).index(x))

    p = sns.barplot(temp, x="year", y="id", hue=hue, ax=ax, order=x_order, hue_order=sorted_months, palette=palette)

    if ax:
        ax.set_title(title)
        ax.legend(loc="upper left")

    return p


def plot_author_associations(data: pd.DataFrame, **kwargs) -> Union[plt.Figure, plt.Axes, pd.DataFrame]:
    ax: Optional[Axes] = kwargs.get("ax")
    palette: Optional[str] = kwargs.get("palette")
    title: Optional[str] = kwargs.get("title", "Author Associations")

    temp = data.groupby("author_association")["id"].count()
    temp.name = "count"

    return temp.plot.pie(ax=ax, title=title, legend=True, colormap=palette)


def plot_conventional_commit_breakdown(data: pd.DataFrame, **kwargs) -> Union[plt.Figure, plt.Axes, pd.DataFrame]:
    ax: Optional[Axes] = kwargs.get("ax")
    palette: Optional[str] = kwargs.get("palette")
    title: Optional[str] = kwargs.get("title", "Conventional Commit Breakdown")
    hue: Optional[str] = kwargs.get("hue", "type")

    conventional_commit_frame = (
        data["title"].str.extract(r"^(?P<type>feat|fix|docs|style|refactor|test|chore|build|ci|perf)(\((?P<scope>[A-Za-z-]+)\))?: (?P<subject>[^\n]+)$").copy()
    )

    conventional_commit_frame = conventional_commit_frame.drop(columns=[1, "subject"])
    conventional_commit_frame = pd.concat([conventional_commit_frame, data["repository_name"]], axis=1)
    conventional_commit_frame = conventional_commit_frame.dropna(subset=["type"])

    if conventional_commit_frame.empty:
        logger.warning("passed data did not seem to have any conventional commits")
        return None

    type_count = conventional_commit_frame.groupby(["type", "repository_name"]).count().reset_index()
    type_count = type_count[type_count["scope"] != 0]
    type_count = type_count.sort_values(by="scope", ascending=False)
    type_count = type_count.rename(columns={"scope": "count"})

    p = sns.barplot(type_count, y="repository_name", x="count", hue=hue, ax=ax, palette=palette)

    if ax:
        ax.set_title(title)
        ax.legend(loc="upper right")

    return p


def plot_reactions(data: pd.DataFrame, **kwargs) -> Union[plt.Figure, plt.Axes, pd.DataFrame]:
    ax: Optional[Axes] = kwargs.get("ax")
    palette: Optional[str] = kwargs.get("palette")
    title: Optional[str] = kwargs.get("title", "Reactions")
    hue: Optional[str] = kwargs.get("hue")
    threshold: Optional[int] = kwargs.get("threshold")

    reaction_columns = [x for x in data.columns.tolist() if x.startswith("reactions.") and x not in ["reactions.url"]]

    reaction_df = data[reaction_columns]
    reaction_df = reaction_df[reaction_df["reactions.total_count"] > 0]
    reaction_df = reaction_df.drop(columns=["reactions.total_count"])

    if reaction_df.empty:
        logger.warning("passed data did not seem to have any reactions 🙁")
        return None

    if threshold:
        reaction_df = reaction_df[reaction_df < threshold]

    p = sns.scatterplot(reaction_df, ax=ax, palette=palette, hue=hue)

    if ax:
        ax.set_title(title)
        ax.legend(loc="upper right")

    return p



import logging
from datetime import datetime
from typing import List, Optional, Tuple

import pandas as pd
from github import Github
from github.GithubException import RateLimitExceededException
from rich.spinner import Spinner

from prfiesta.environment import GitHubEnvironment
from prfiesta.spinner import update_spinner

logger = logging.getLogger(__name__)


class GitHubCollector:
    def __init__(self, **kwargs) -> None:
        environment = GitHubEnvironment()
        token = kwargs.get("token") or environment.get_token()
        self._url = kwargs.get("url") or environment.get_url()

        self._github = Github(token, base_url=self._url)
        self._spinner: Spinner = kwargs.get("spinner")

        self._sort_column = ["updated_at"]
        self._drop_columns = kwargs.get("drop_columns") or ["node_id", "performed_via_github_app"]

        self._move_to_end_columns = [
            "url",
            "repository_url",
            "html_url",
            "timeline_url",
            "labels_url",
            "comments_url",
            "events_url",
        ]
        self._datetime_columns = [
            "created_at",
            "updated_at",
            "closed_at",
        ]

    def collect(
        self,
        *users: Optional[Tuple[str]],
        after: Optional[datetime] = None,
        before: Optional[datetime] = None,
        use_updated: Optional[bool] = False,
        use_involves: Optional[bool] = False,
        use_reviewed_by: Optional[bool] = False,
        use_review_requested: Optional[bool] = False,
        reference: Optional[str] = None,
    ) -> pd.DataFrame:
        query = self._construct_query(users, after, before, use_updated, use_involves, use_reviewed_by, use_review_requested, reference)

        update_spinner(f"Searching {self._url} with[bold blue] {query}", self._spinner, logger)

        pull_request_data = None
        try:
            pulls = self._github.search_issues(query=query)

            pull_request_data: list[dict] = []
            for pr in pulls:
                pull_request_data.append(pr.__dict__["_rawData"])

        except RateLimitExceededException as e:
            logger.warning("🙇 You were rate limited by the GitHub API, try requesting less data.")
            logger.debug(e)
            return pd.DataFrame()

        if not pull_request_data:
            logger.warning("Did not find any results for this search criteria!")
            return pd.DataFrame()

        update_spinner("Post Processing", self._spinner, logger)
        pr_frame = pd.json_normalize(pull_request_data)

        pr_frame = pr_frame.drop(columns=self._drop_columns, errors="ignore")
        pr_frame = pr_frame.sort_values(by=self._sort_column, ascending=False)
        pr_frame = self._parse_datetime_columns(pr_frame)
        pr_frame["repository_name"] = pr_frame["repository_url"].str.extract(r"(.*)\/repos\/(?P<repository_name>(.*))")["repository_name"]
        pr_frame = self._move_column_to_end(pr_frame)

        return pr_frame

    @staticmethod
    def _construct_query(
        users: Optional[List[str]],
        after: Optional[datetime] = None,
        before: Optional[datetime] = None,
        use_updated: Optional[bool] = False,
        use_involves: Optional[bool] = False,
        use_reviewed_by: Optional[bool] = False,
        use_review_requested: Optional[bool] = False,
        reference: Optional[str] = None,
    ) -> str:
        """
        Constructs a GitHub Search Query
        that returns pull requests made by the passed users and options.

        Examples
        --------
            type:pr author:user1
            type:pr author:user2 created:<=2021-01-01
            type:pr author:user1 author:user2 created:2021-01-01..2021-03-01
            type:pr author:user2 updated:>=2021-01-01
            type:pr involves:user2
            type:pr reviewed-by:user1
            type:pr review-requested:user1
            type:pr in:title,body "PA-12765"

        All dates are inclusive.
        See GitHub Docs for full options https://docs.github.com/en/search-github/searching-on-github/searching-issues-and-pull-requests
        """
        query: List[str] = []
        query.append("type:pr")

        author_filter = "author"
        if use_involves:
            author_filter = "involves"
        elif use_reviewed_by:
            author_filter = "reviewed-by"
        elif use_review_requested:
            author_filter = "review-requested"

        logger.debug("using author filter %s", author_filter)

        for u in users:
            query.append(f"{author_filter}:{u}")

        time_filter = "created"
        if use_updated:
            time_filter = "updated"

        logger.debug("using time filter %s", time_filter)

        if before and after:
            query.append(f"{time_filter}:{after.strftime('%Y-%m-%d')}..{before.strftime('%Y-%m-%d')}")
        elif before:
            query.append(f"{time_filter}:<={before.strftime('%Y-%m-%d')}")
        elif after:
            query.append(f"{time_filter}:>={after.strftime('%Y-%m-%d')}")

        if reference:
            query.append(f'in:title,body "{reference}"')

        return " ".join(query)

    def _move_column_to_end(self, df: pd.DataFrame) -> pd.DataFrame:
        for col in self._move_to_end_columns:
            try:
                df.insert(len(df.columns) - 1, col, df.pop(col))
                df.drop(columns=col)
            except KeyError:
                # This can happen if the user provides a custom _drop_columns which
                # removes the column before we can move it to the end
                logger.debug("Attempted to move column %s but it did not exist", col)

        return df

    def _parse_datetime_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        for col in self._datetime_columns:
            df[col] = pd.to_datetime(df[col], errors="coerce")
        return df


if __name__ == "__main__":  # pragma: nocover
    g = GitHubCollector()
    logger.info(g._construct_query(["kiran94", "hello"], datetime(2021, 1, 1), datetime(2021, 3, 1)))
